#!/usr/bin/env perl
use warnings;
use strict;
use diagnostics;

# Global variables ------------------
my @pc_violations = (); # pc of isolated anomalies
my %violations =();	# a hash of the anomalies, indexed by pc
my @validated = (); # pc's of validated anomalies
my @unresolved = (); # pc's of unresolved anomalies 
					 #(e.g. crash will different stack trace)
my @dismissed =(); # anomalies dismissed as false-positives
my $run_command; # command used to run the tool/application
my $input_file; # isolated anomalies filename
my $output_file; # validation output filename
my $delete_outcome_file = "DeleteOut.txt"; 
my $bad_st_filename; # filename for bad stacktrace
my @bad_trace =();
my $good_st_filename; # filename for good stacktrace
my @good_trace =();
my $log; 	# execution log

my $split_chunk = 1; # in the case of splitting the code into regions
# -----------------------------------


# Function prototypes ---------------
sub check_crash($);
sub read_stack_trace($);
sub read_stack_trace2($);
sub compare_stack_traces;
sub read_violations();
sub validate(\@);
# -----------------------------------


# -------------------------------------------------------------
# Main program 
# -------------------------------------------------------------
$run_command = shift(@ARGV) or die "Please supply a run command name\n";
$input_file = shift(@ARGV) or die "Please supply an input file name\n";
$output_file = shift(@ARGV)or die "Please supply a output file name\n";
#$bad_st_filename = shift(@ARGV) or die "Please supply a fail stack trace\n";
$bad_st_filename = shift(@ARGV);
$split_chunk = shift(@ARGV);

if(! defined($bad_st_filename)){
	print "\n\n". "*" x 40 . "\n";
	print "Crash stacktrace was not specified.\n";
	print "Running the program to extract the crash stacktrace...\n";
	print "*" x 40 . "\n\n\n";
	
	my $cmd = "./$run_command 1 0 $delete_outcome_file";
	system($cmd) == 0 or die "Error: could not '$cmd'";

	#$bad_trace = read_stack_trace($delete_outcome_file);
	@bad_trace = read_stack_trace2($delete_outcome_file);
	open (my $fh, "> bad_stack_trace.txt");
	foreach my $elem (@bad_trace){
		print $fh $elem;
	}
}
else{
	print "\n\n". "*" x 40 . "\n";
	print "Crash stacktrace specified in file: $bad_st_filename\n";
	print "Reading in the crash stacktrace...\n";
	print "*" x 40 . "\n\n\n";
	
	#$bad_trace = read_stack_trace($bad_st_filename);
	@bad_trace = read_stack_trace2($bad_st_filename);
	
}

# Read-in the violations produced by anomaly detection
read_violations();

validate(@pc_violations);

print "\n\n". "*" x 40 . "\n";
print "The validated anomalies are in file $output_file\n";
print "*" x 40 . "\n\n\n";

open (my $fh, ">$output_file");
print $fh "*" x 40 . "\n";
print $fh "SUMMARY\n";
print $fh "*" x 40 . "\n\n";
print $fh "Validated anomalies\n";
print $fh "*" x 40 . "\n";
foreach (@validated) {
	print $fh "$violations{$_}";
}
print $fh "\nUnresolved anomalies\n";
print $fh "*" x 40 . "\n";
foreach (@unresolved) {
	print $fh "$violations{$_}";
}
print $fh "\nDismissed anomalies\n";
print $fh "*" x 40 . "\n";
foreach (@dismissed) {
	print $fh "$violations{$_}";
}
print $fh "\n\n" . "*" x 40 . "\n";
print $fh "Detailed Log\n";
print $fh "*" x 40 . "\n";
print $fh "$log";

close($fh);
exit 0;
# -------------------------------------------------------------

# Check to see if the application completed with a crash or not 
# Return true if crashed, otherwise return false
sub check_crash($) {

	my $filename = shift;
	open(INFILE, $filename)   or die "Can't open $filename: $!";
	my $app_crashed = <INFILE>; # read the first line
	close INFILE;

	return ( $app_crashed =~ /Application Crashed: YES/i);
}

# Read in the stacktrace part of the Detele output file
sub read_stack_trace($){

	my $filename = shift;
	my $start_stacktrace = 0;
	my $stacktrace;
	open(INFILE, $filename)   or die "Can't open $filename: $!";
	
	while(<INFILE>){
		if(/Backtrace:/){
			$start_stacktrace = 1; # flag to start reading the stacktrace
		}
		elsif($start_stacktrace == 1){
			$stacktrace = $stacktrace ."$_";				
		}
	}
	close INFILE;

	($start_stacktrace == 1) or die "Bad file. Did not extract stacktrace!! $!\n";
	return $stacktrace;
}

sub read_stack_trace2($){

	my $filename = shift;
	my $start_stacktrace = 0;
	my @stacktrace;
	open(INFILE, $filename)   or die "Can't open $filename: $!";
	
	while(<INFILE>){
		if(/Backtrace:/){
			$start_stacktrace = 1; # flag to start reading the stacktrace
		}
		elsif($start_stacktrace == 1){
			push @stacktrace, $_;
		}
	}
	close INFILE;

	($start_stacktrace == 1) or die "Bad file. Did not extract stacktrace!! $!\n";
	return @stacktrace;
}

# return 0 if stack traces are differnt, else return 1
#sub compare_stack_traces{
#	my ($first, $second) = @_;
#	for (my $i = 0; $i < @$first; $i++) {
#		return 0 if $first->[$i] ne $second->[$i];
#	}
#	return 1;
#}


sub compare_stack_traces{
	my ($first, $second) = @_;
	my $equal = 0;
	my $different = 0;

	# move one of the arrays into a hash
	my %count = ();
	for (@$first) { $count{$_} = 1 }

	# iterate over the second array and 
	# count the number of elements that are in the hash

	for(@$second){
		if( exists($count{$_})){
			$equal++;	
		}
		else{
			$different++;
		}
	}

	if($different > 2){
		return 0; # Too much different
	}
	else{
		return 1; # Consider similar
	}	
}

# Read the output file with violations 
sub read_violations() {
	
	open(INFILE, "$input_file") 
		or die "Can't open $input_file: $!";

	# read every line from input file
	while(<INFILE>){
		
		# find the pc in the input line
		# convert the pc to decimal and store it as a key of a hash. 
		# The whole violation line is the value of the hash. 
		# NOTE: this will automatically remove violations from the 
		# same pc and will keep only the last one. 
#		if(	/pc:\s([1234567890abcdef]+)/){ 
		if(	/([1234567890abcdef]+)/){ 
			$violations{ hex($1) } = $_;
		}
		else{
			print "ERROR: bad format of input file. Did not find pc: \n";
		}
	}

	# Get the pc's of violations into an array in sorted order
	@pc_violations = sort {$a<=>$b} ( keys %violations );
}


# -------------------------------------------------------------
# Isolate: isolate the relevant violations, 
# 		   i.e those that carry a token at the end of execution
# -------------------------------------------------------------
sub validate(\@){
	
	my $array_ref = shift;
	# copy the whole input array into @half1
	my @array = @$array_ref; 

	# get the number of elements of the array
	my $num_elem = @array; 

	my $msg; # save detailed messages to log

	print "\n\n". "*" x 40 . "\n";
	print "Starting validation by automatically nullifying anomalous instructions\n";
	print "*" x 40 . "\n";
		

	foreach (@array){

		my $anomaly = $_;
		print "\n\n". "*" x 40 . "\n";
		$msg = "Nullifying anomaly: $violations{$anomaly}";	
		print $msg;
		$log = $log . "\n\n$msg";
		print "*" x 40 . "\n\n\n";

		#my $cmd = "./$run_command 1 $_ $delete_outcome_file";
		my $cmd = "./$run_command 1 $_ $delete_outcome_file $split_chunk";
#		system($cmd) == 0 or die "Error: could not '$cmd'";
		system($cmd);

		#my $temp_stacktrace = read_stack_trace($delete_outcome_file);
		my @temp_stacktrace = read_stack_trace2($delete_outcome_file);

		my $bad_trace = @bad_trace; # number of elements in the bad trace
		my $temp_stacktrace = @temp_stacktrace; # number of elements in the temp trace

		# The application crashed. 
		# Check the stack trace to see if it is the same as
		# out crash stacktrace. If it is, then we assume that the
		# instruction is not relevant for the crash. If the
		# stack trace is different, then the outcome is unresoved. 
		if(check_crash($delete_outcome_file)){
			$log = $log . "Application crash: YES\n";
			if( ($bad_trace > 0 ) and ($temp_stacktrace > 0)){
				if( compare_stack_traces( \@bad_trace, \@temp_stacktrace)){
					print "\n\n". "*" x 40 . "\n";
					$msg = "Stack traces match.\n";
					$msg = $msg . "Anomaly: $violations{$anomaly} classified as false-positive\n";
					print $msg;
					$log = $log . "$msg";
					print "*" x 40 . "\n\n\n";
					push @dismissed, $anomaly;
				}
				else{
					print "\n\n". "*" x 40 . "\n";
					$msg = "Stack traces do not match.\n";
					$msg = $msg . "Anomaly: $violations{$anomaly} classified as unknown\n";
					print $msg;
					$log = $log . "$msg";
					print "*" x 40 . "\n\n\n";
					push @unresolved, $anomaly;
				}
			}
			else{
				print "\n\n". "*" x 40 . "\n";
				$msg = "Unable to compare stack traces!\n";
				$msg = $msg . "Anomaly: $violations{$anomaly} classified as false-positive\n";
				print $msg;
				$log = $log . "$msg";
				print "*" x 40 . "\n\n\n";
				push @dismissed, $anomaly;
			}
			
		} #check crash
		# check crash is false, the application run passed
		# Compare stacktraces. 
		else{
			$log = $log . "Application crash: NO\n";
			print "\n\n". "*" x 40 . "\n";
			$msg = "We are not comparing passing stack traces!\n";
			$msg = $msg . "Anomaly: $violations{$anomaly} classified as relevant\n";
			print $msg;
			$log = $log . "$msg";
			print "*" x 40 . "\n\n\n";
			push @validated, $anomaly;
	
			# If the application did not crash, save the output for comparisson
		#	my $mv_command = "mv ../outputs/fail.s ../outputs/".$violations{$anomaly};
		#	print $mv_command . "\n";
		#	system $mv_command;
			
			$log = $log . "Stacktrace:\n";
			foreach my $item (@temp_stacktrace){
				$log = $log . " " . $item;
			}

		}# else
	}# for each
}# verify


